/* pathnames.c */

#include <Errors.h>

#include "MoreFilesExtras.h"

#include "DGDebugging.h"
#include "pstring.h"
#include "NGLList.hh"

#include "SiesString.hh"

#include "pathnames.hh"

char Char2Nibble(char c);

char Char2Nibble(char c)
{
	return (c >= 'A') ? (c | ('a' - 'A') - 'a' + 10) : c - '0';
}

OSErr MangleMacFilename(const unsigned char *inName, unsigned char *outName)
{
	const unsigned char *p;
	unsigned char *q;
	
	// Check for all-dot names first
	if (inName[1] == '.') {
		bool done = false;
		for (p = inName + inName[0]; *p == '.' && (p > inName + 1 || (done = true) && 0); p--)
			;
		if (done) {
			PstrPcpy(outName, inName);
			PstrPcat(outName, "\p..");
			return 0;
		}
	}
	for (p = inName + 1, q = outName + 1; p <= inName + inName[0]; p++, q++) {
		if (0 && *p == ' ') {
			*q++ = '\\';
			*q = *p;
		} else if (*p == '/') {
			*q++ = ':';
			*q++ = '2';
			*q = 'f';
		} else
			*q = *p;
	}
	outName[0] = q - 1 - outName;
	return 0;
}

OSErr UnmangleMacFilename(const unsigned char *inName, unsigned char *outName)
{
	const unsigned char *p;
	unsigned char *q;
	
	// Check for all-dot names first
	if (inName[0] > 2 && inName[1] == '.') {
		bool done = false;
		for (p = inName + inName[0]; *p == '.' && (p > inName + 1 || (done = true) && 0); p--)
			;
		if (done) {
			PstrPcpy(outName, inName);
			outName[0] -= 2;
			return 0;
		}
	}
	for (p = inName + 1, q = outName + 1; p <= inName + inName[0]; p++, q++) {
		if (*p == '\\') {
			if (p + 1 > inName + inName[0])
				break;
			*q = *++p;
		} else if (*p == ':') {
			char c;
			if (p + 2 > inName + inName[0])
				break;
			c = Char2Nibble(*++p) << 4;
			c += Char2Nibble(*++p);
			*q = c;
		} else
			*q = *p;
	}
	outName[0] = q - 1 - outName;
	return 0;
}

OSErr UnixPathname2DirAndName(
	const FSDirSpec &inCWD, 
	const char *inPathname, 
	FSDirSpec &outDir, 
	SiesString &outFilename)
{
	//*outFilename = inPathname;
	outDir = inCWD;
	
	if (inPathname == NULL)
		return paramErr;
	
	if (inPathname[0] == '\0')
		return paramErr;
	
	// We don't support unix absolute pathnames yet.
	// Use a home-dir pathname instead.
	if (inPathname[0] == '/')
		return dirNFErr;
		
	return UnixRelativePathname2DirAndName(inCWD, inPathname, outDir, outFilename);
}

OSErr UnixRelativePathname2DirAndName(
	const FSDirSpec &inCWD, 
	const char *inPathname, 
	FSDirSpec &outDir, 
	SiesString &outFilename)
{
	Str255 dirName, homeName, scratch;
	const char *p, *q;
	FSDirSpec dir;
	long dirID;
	OSErr err;
	Boolean isDir;
	
	//*outFilename = inPathname;
	outDir = inCWD;
	
	if (inPathname == NULL)
		return paramErr;
	
	if (inPathname[0] == '\0')
		return paramErr;
	
	if (inPathname[0] == '/')
		return dirNFErr;
	
	dir = inCWD;
	isDir = true;
	p = q = inPathname;
	while (*q != '\0') {
		// scan to the first/next slash
		while ((*q != '\0') && (*q != '/'))
			q++;
		// copy the first/next item name -- null string if absolute pathname
		Pstrncpy(dirName, p, q - p);
		// if not an absolute pathname
		if (q - p > 0) {
			// check the first char
			switch (p[0]) {
#if 1
				case '~':  // home dir -> named volume
					p++;
					if (q == p) {
						// q == p -> ~ is followed by /  or end of string.
						// Therefore, it means 'my home directory' on unix.
						// Get the first drive in the queue (the startup disk?)
						err = DetermineVRefNum(NULL, 1, &dir.vRefNum);
					} else {
						// Otherwise, it's someone else's home dir.
						// For us, it's a named volume.  Copy the name.
						Pstrncpy(homeName, p, q - p);
						UnmangleMacFilename(homeName, scratch);
						PstrPcat(scratch, "\p:");
						err = DetermineVRefNum(scratch, 0, &dir.vRefNum);
					}
					if (err != noErr)
						return dirNFErr;
					dir.dirID = 2; // root directory
					break;
#endif
				case '.':  // current dir, parent dir, or dotfile
					if (q - p == 1) {
						// current dir; no change
						break;
					} else if (q - p == 2 && p[1] == '.') {
						// parent dir
						err = GetParentID(dir.vRefNum, dir.dirID, NULL, &dirID);
						if (err != noErr)
							return err;
						dir.dirID = dirID;
						break;
					}
					// Otherwise, not . or ..
					/* fall thru */
				default:  // 
					UnmangleMacFilename(dirName, scratch);
					err = GetDirectoryID(dir.vRefNum, dir.dirID, scratch, &dirID, &isDir);
					if (err != noErr || !isDir) {
						// If there's no directory, it's a file (possibly nonexistent)
						outFilename.Set(p);
						outDir = dir;
						// If we require a directory but it wasn't found...
						if (err == fnfErr && *q == '/')
							err = dirNFErr;
						return err;
					}
					dir.dirID = dirID;
					break;
			}
		}
		if (*q != '\0')
			p = ++q;
	}
	outDir = dir;
	outFilename.Set(q);
	return noErr;
}

OSErr
DirAndName2UnixAbsolutePathname(const FSDirSpec *inDir, const char *inFilename, Handle ioPathname)
{
	OSErr err;
	FSDirSpec dir;
	long dirID;
	long len;
	char *p;
	NGLList<SiesString *> list;
	short count = 0;
	
	len = 2 + (inFilename ? strlen(inFilename) : 0); // "/" + '\0'
	dir = *inDir;
	do {
		// get dirname
		Str255 name;
		err = GetDirName(dir.vRefNum, dir.dirID, name);
		ThrowIfOSErr_(err);
		len += name[0] + 1;
		// save dirname
		SiesString *str = new SiesString(name);
		list.Prepend(str);
		count++;
		err = GetParentID(dir.vRefNum, dir.dirID, NULL, &dirID);
		ThrowIfOSErr_(err);
		dir.dirID = dirID;
	} while (dirID != 1);
	if (err != noErr) {
		return err;
	}
	
	SetHandleSize(ioPathname, len);
	if (err = MemError())
		return err;
	short hState = HGetState(ioPathname);
	HLock(ioPathname);
	p = *ioPathname; // Dereferenced Handle
	p[0] = '/';
	//p[1] = '\0';
	p++;
	for (; count > 0; count--) {
		SiesString *str = list.Behead();
		// copy dirname onto p
		Str255 name;
		str->GetPStr(name, 255);
		delete str;
		strPcpy(p, name);
		// update p
		p += name[0];
		p[0] = '/';
		p++;
	}
	HSetState(ioPathname, hState);
	if (inFilename) {
		strcpy(p, inFilename);
		p += strlen(inFilename);
	}
	p[0] = '\0';
	
	return noErr;
}
